/*-
 * Copyright (C)2004 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

struct eoldetector {
    ssize_t eolp;
    int eolst, eomax;
    const char *eoln;
    const char * const * buf;
    ssize_t eotrix[1];
};
/******************************************/
/* eomax == -1 eoln    */
/* eomax >= 0  eoln                  */
/* eolst == -1 eoln       */
/* eolst >= 0           */
/******************************************/
static struct eoldetector *
eolinit(u_int32_t flag, const char *eoln, const char * const * buf) {
    struct eoldetector *eo;
    int i;

    ifBLIN_QV3(flag) fprintf( stderr, "+eolinit=%s~\n", eoln);
    eo = NULL;
    if  (eoln) {
        eo = malloc(strlen(eoln) * sizeof(ssize_t) + sizeof(struct eoldetector));
        if  (!eo) {
            errno = ENOMEM;
            ifBLIN_QV5(flag) warn("eolinit");
            ifBLIN_QV3(flag) fprintf( stderr, "-eolinit NULL\n");
            return(NULL);
        }
        eo->buf = buf;
        eo->eomax = strlen(eoln);
        eo->eolst = -1;
        eo->eolp = 0;
        eo->eoln = eoln;
        for (i = 0; i < eo->eomax; i++) {
            if  (i == 0) (eo->eotrix)[0] = 0;
            else if  (eoln[(eo->eotrix)[i - 1]] != eoln[i]) (eo->eotrix)[i] = 0;
            else (eo->eotrix)[i] = (eo->eotrix)[i - 1] + 1;
            ifBLIN_QV4(flag) fprintf( stderr, " %08X", (eo->eotrix)[i]);
    }   }
    ifBLIN_QV3(flag) fprintf( stderr, "\n-eolinit %08X\n", (u_int32_t)eo);
    return(eo);
}

static void
eolstop(struct eoldetector *eo) {
    if  (eo) eo->eolst = -1;
}

static void
eolstart(struct eoldetector *eo, ssize_t eolp) {
    if  (eo && eo->eolst <= 0) {
        eo->eolp = eolp;
        eo->eolst = 0;
}   }

static void
eoldelta(struct eoldetector *eo, off_t delta) {
    if  (eo && eo->eolst >= 0) {
        eo->eolp += delta;
        if  (eo->eolp < 0) {
            eo->eolp = 0;
            eo->eolst = 0;
}   }   }

static int
eoldetect(u_int32_t flag, struct eoldetector *eo, ssize_t len) {
    int r;

    ifBLIN_QV3(flag) fprintf( stderr, "+eoldetect %d\n", len);
    r = -1;
    if  (eo && eo->eolst >= 0) {
        while (eo->eolp <= len - eo->eomax && r < 0) {
            ifBLIN_QV4(flag) fprintf(stderr, "eolp=%d eolst=%d eomax=%d\n", eo->eolp, eo->eolst, eo->eomax);
            if (eo->eolst >= eo->eomax) {
                r = eo->eolp;
            } else if ((*eo->buf)[eo->eolp++] != (eo->eoln)[eo->eolst++]) {
                if  (--eo->eolst) --eo->eolp;
                eo->eolst = (eo->eotrix)[eo->eolst];
    }   }   }
    ifBLIN_QV3(flag) fprintf( stderr, "-eoldetect %d\n", r);
    return(r);
}
